!
! Simple (and inefficient!) molecular dynamics code for a Lennard-Jones system.
! The code is only meant for pedagogical purposes, not for production.
! Author: Daniele Coslovich (daniele.coslovich@umontpellier.fr)
!
MODULE kernels

  IMPLICIT NONE

  INTEGER, PARAMETER :: NDIM = 3
  INTEGER, PARAMETER :: DP = SELECTED_REAL_KIND(12)

CONTAINS

  !! Read a configuration in xyz format from file.
  !! Return: box size, positions and velocities of particles
  SUBROUTINE read(unit,box,pos,vel)
    INTEGER, INTENT(in)  :: unit
    REAL(DP),ALLOCATABLE :: box(:)
    REAL(DP),ALLOCATABLE :: pos(:,:), vel(:,:)
    INTEGER              :: i, n
    CHARACTER(6)         :: string,tmp
    READ(unit,*) n
    IF (ALLOCATED(pos)) DEALLOCATE(pos)
    IF (ALLOCATED(vel)) DEALLOCATE(vel)
    IF (ALLOCATED(box)) DEALLOCATE(box)
    ALLOCATE(pos(NDIM,n),vel(NDIM,n),box(NDIM))
    READ(unit,*) box
    DO i = 1,n
       READ(unit,*) string,pos(:,i),vel(:,i)
    END DO
  END SUBROUTINE read

  !! Write a configuration in xyz format
  SUBROUTINE write(unit,box,pos,vel)
    INTEGER,  INTENT(in) :: unit
    REAL(DP), INTENT(in) :: box(:)
    REAL(DP), INTENT(in) :: pos(:,:), vel(:,:)
    INTEGER              :: i
    WRITE(unit,*) SIZE(pos,2)
    WRITE(unit,*) box
    DO i = 1,SIZE(pos,2)
       WRITE(unit,"(a,6f12.5)") "A ",pos(:,i), vel(:,i)
    END DO
  END SUBROUTINE write

  !! Read simulation parameters from input file provided on the command line
  SUBROUTINE input(input_cnf,output_cnf,output_log,nsteps,dt,period_cnf,period_log)
    CHARACTER(256), INTENT(inout) :: input_cnf,output_cnf,output_log
    INTEGER,        INTENT(inout) :: nsteps,period_cnf,period_log
    REAL(DP),       INTENT(inout) :: dt
    INTEGER                       :: unit=99
    CHARACTER(256)                :: input_file, param, value
    IF (COMMAND_ARGUMENT_COUNT() /= 1) THEN
       WRITE(*,"(a)") "Usage: ./a.out <parameter_file>"
       WRITE(*,"(a)") "where <parameter_file> is a file with the simulation parameters."
       STOP 
    END IF
    CALL GET_COMMAND_ARGUMENT(1,value=input_file)
    OPEN(unit=unit,file=input_file,status="old")
    DO
       READ(unit,*,END=10) param, value
       IF (TRIM(param) == "period_configs") READ(value(1:LEN_TRIM(value)),*) period_cnf
       IF (TRIM(param) == "period_log") READ(value(1:LEN_TRIM(value)),*) period_log
       IF (TRIM(param) == "file_config") READ(value(1:LEN_TRIM(value)),*) output_cnf
       IF (TRIM(param) == "file_log") READ(value(1:LEN_TRIM(value)),*) output_log
       IF (TRIM(param) == "file_input") READ(value(1:LEN_TRIM(value)),*) input_cnf
       IF (TRIM(param) == "nsteps") READ(value(1:LEN_TRIM(value)),*) nsteps
       IF (TRIM(param) == "dt") READ(value(1:LEN_TRIM(value)),*) dt
    END DO
10  CONTINUE
  END SUBROUTINE input

  !! Integration step using the velocity-Verlet algorithm
  SUBROUTINE evolve(dt,rcut,box,pos,pos_unf,vel,for,epot)
    REAL(DP), INTENT(in)    :: dt, rcut
    REAL(DP), INTENT(in)    :: box(:)
    REAL(DP), INTENT(inout) :: pos(:,:), pos_unf(:,:), vel(:,:), for(:,:)
    REAL(DP), INTENT(out)   :: epot
    REAL(DP), PARAMETER     :: mass = 1.0
    INTEGER                 :: i
    pos = pos + vel * dt + 0.5 * for / mass * dt**2
    pos_unf = pos_unf + vel * dt + 0.5 * for / mass * dt**2
    vel = vel + 0.5 * for / mass * dt
    DO i = 1,SIZE(pos,2) 
       CALL pbc(pos(:,i),box)
    END DO
    CALL forces(rcut,box,pos,for,epot)
    vel = vel + 0.5 * for / mass * dt
  END SUBROUTINE evolve

  !! Dump some thermodynamic quantities to the specified IO unit
  SUBROUTINE report(what,unit,step,box,pos,vel,epot)
    CHARACTER(*), INTENT(in) :: what
    INTEGER,      INTENT(in) :: unit, step
    REAL(DP),     INTENT(in) :: pos(:,:), vel(:,:), box(:), epot
    REAL(DP)                 :: ekin
    INTEGER                  :: ndim, npart
    IF (what == "header") THEN
       WRITE(unit,"('#',a9,4a14)") "Step", "Epot", "Temp", "Ekin", "Etot"
       WRITE(unit,"('#',65('-'))")
    ELSE IF (what == "log") THEN
       ekin = kinetic(vel)
       ndim = SIZE(pos,1)
       npart = SIZE(pos,2)
       WRITE(unit,"(i10,4es14.6)") step, epot, 2.0*ekin/(npart*ndim), ekin, epot+ekin
    END IF
  END SUBROUTINE report

  !! Evaluate the kinetic energy from the particles' velocities
  FUNCTION kinetic(vel) RESULT (ekin)
    REAL(DP), INTENT(in) :: vel(:,:)
    REAL(DP)             :: ekin
    INTEGER              :: i
    ekin = 0.0
    DO i = 1,SIZE(vel,2)
       ekin = ekin + DOT_PRODUCT(vel(:,i),vel(:,i))
    END DO
    ekin = 0.5 * ekin
  END FUNCTION kinetic

  !! Evaluate the potential energy and the virial contribution 
  !! using the cut and shifted Lennard-Jones potential
  SUBROUTINE potential(rcut,rsq,u,w)
    REAL(DP), INTENT(in)  :: rsq, rcut
    REAL(DP), INTENT(out) :: u, w
    REAL(DP)              :: uc
    uc = 4.0/rcut**12 - 4.0/rcut**6
    u = 4.0 * (1.0/rsq**6 - 1.0/rsq**3) - uc
    w = 24.0 * (2.0/rsq**6 - 1.0/rsq**3 ) / rsq
  END SUBROUTINE potential

  !! Apply minimum image convention to a distance vector r.
  !! This subroutine can also be used to fold a particle back 
  !! in the central cell during a simulation. However, it won't
  !! correctly fold back a particle from an arbitrary position.
  SUBROUTINE pbc(r,box)
    REAL(DP), INTENT(inout) :: r(:)
    REAL(DP), INTENT(in)    :: box(:)
    WHERE (ABS(r) > box/2)
       r = r - SIGN(box,r)
    END WHERE
  END SUBROUTINE pbc

  !! Evaluate the potential energy and the forces between particles
  SUBROUTINE forces(rcut,box,pos,for,epot)
    REAL(DP), INTENT(in)  :: box(:), rcut
    REAL(DP), INTENT(in)  :: pos(:,:)
    REAL(DP), INTENT(out) :: for(:,:)
    REAL(DP), INTENT(out) :: epot
    REAL(DP)              :: rij(SIZE(pos,1)), rijsq, uij, wij
    INTEGER               :: i, j
    for = 0.0
    epot = 0.0
    DO i = 1,SIZE(pos,2)
       DO j = i+1,SIZE(pos,2)
          rij = pos(:,i) - pos(:,j)
          CALL pbc(rij,box)
          rijsq = DOT_PRODUCT(rij,rij)
          IF (rijsq < rcut**2) THEN
             CALL potential(rcut,rijsq,uij,wij) ! wij = -(du/dr)/r
             epot = epot + uij
             for(:,i) = for(:,i) + wij * rij
             for(:,j) = for(:,j) - wij * rij
          END IF
       END DO
    END DO
  END SUBROUTINE forces

END MODULE kernels

PROGRAM main

  USE kernels

  IMPLICIT NONE

  REAL(DP),ALLOCATABLE :: pos(:,:), pos_unf(:,:), vel(:,:), for(:,:), box(:)
  INTEGER              :: unit_inp=100, unit_log=101, unit_cnf=102
  INTEGER              :: i, nsteps=1000, period_log=10, period_cnf=10
  CHARACTER(256)       :: file_inp="input.xyz", file_log="output.dat", file_cnf="output.xyz"
  REAL(DP)             :: dt=0.002, epot, rcut=2.5, dum

  CALL input(file_inp,file_cnf,file_log,nsteps,dt,period_cnf,period_log)
  OPEN(unit=unit_inp,file=file_inp,status="old")
  OPEN(unit=unit_log,file=file_log,status="unknown")
  OPEN(unit=unit_cnf,file=file_cnf,status="unknown")

  ! Read input configuration and allocate arrays
  CALL read(unit_inp,box,pos,vel)
  ALLOCATE(for(SIZE(pos,1),SIZE(pos,2)))
  ALLOCATE(pos_unf(SIZE(pos,1),SIZE(pos,2)))
  pos_unf = pos

  ! Dumps to log file
  WRITE(unit_log,"('# Number of particles =',i15)") SIZE(pos,2)
  WRITE(unit_log,"('# Box side            =',f15.9)") box(1)
  WRITE(unit_log,"('# Time step           =',es15.3)") dt
  WRITE(unit_log,"('#')") 
  CALL report("header",unit_log,i,box,pos,vel,epot)

  ! Main MD loop
  CALL forces(rcut,box,pos,for,epot)
  DO i = 1,nsteps
     CALL evolve(dt,rcut,box,pos,pos_unf,vel,for,epot)
     IF (MOD(i,period_log)==0) CALL report("log",unit_log,i,box,pos,vel,epot)
     IF (MOD(i,period_cnf)==0) CALL write(unit_cnf,box,pos_unf,vel)
  END DO

  CLOSE(unit_inp)
  CLOSE(unit_cnf)
  CLOSE(unit_log)

END PROGRAM main
